@@ -8,23 +8,17 @@ use warnings NONFATAL => 'all';
 use Protocol::WebSocket::Frame;
 use Scalar::Util qw( weaken );
 use Encode ();
+use AnyEvent;
 use AnyEvent::WebSocket::Message;
 use Carp qw( croak carp );
 
 # ABSTRACT: WebSocket connection for AnyEvent
-our $VERSION = '0.16'; # VERSION
+our $VERSION = '0.17'; # VERSION
 
 
-has _stream => (
-  is => 'ro',
-  required => 1,
-);
-
-has _handle => (
+has handle => (
   is       => 'ro',
-  lazy     => 1,
-  default  => sub { shift->_stream->handle },
-  weak_ref => 1,
+  required => 1,
 );
 
 foreach my $type (qw( each_message next_message finish ))
@@ -43,12 +37,12 @@ sub BUILD
   my $finish = sub {
     $_->($self) for @{ $self->_finish_cb };
   };
-  $self->_handle->on_error($finish);
-  $self->_handle->on_eof($finish);
+  $self->handle->on_error($finish);
+  $self->handle->on_eof($finish);
 
   my $frame = Protocol::WebSocket::Frame->new;
-  
-  $self->_stream->read_cb(sub {
+
+  my $read_cb = sub {
     $frame->append($_[0]{rbuf});
     while(defined(my $body = $frame->next_bytes))
     {
@@ -64,6 +58,21 @@ sub BUILD
         $_->($self, $message) for @{ $self->_each_message_cb };
       }
     }
+  };
+
+  # Delay setting on_read callback. This is necessary to make sure all
+  # received data are handled by each_message and/or next_message
+  # callbacks. If there is some data in rbuf, changing the on_read
+  # callback makes the callback fire, but there is of course no
+  # each_message/next_message callback to receive the message yet.
+  $self->handle->on_read(undef);
+  my $idle_w; $idle_w = AnyEvent->idle(cb => sub {
+    undef $idle_w;
+    if(defined($self))
+    {
+      $read_cb->($self->handle); # make sure to read remaining data in rbuf.
+      $self->handle->on_read($read_cb);
+    }
   });
 }
 
@@ -82,7 +91,7 @@ sub send
   {
     $frame = Protocol::WebSocket::Frame->new($message);
   }
-  $self->_handle->push_write($frame->to_bytes);
+  $self->handle->push_write($frame->to_bytes);
   $self;
 }
 
@@ -115,11 +124,12 @@ sub close
 {
   my($self) = @_;
 
-  $self->_handle->push_write(Protocol::WebSocket::Frame->new(type => 'close')->to_bytes);
-  $self->_handle->push_shutdown;
+  $self->handle->push_write(Protocol::WebSocket::Frame->new(type => 'close')->to_bytes);
+  $self->handle->push_shutdown;
 }
 
 
+
 sub on_each_message
 {
   my($self, $cb) = @_;
@@ -162,7 +172,7 @@ AnyEvent::WebSocket::Connection - WebSocket connection for AnyEvent
 
 =head1 VERSION
 
-version 0.16
+version 0.17
 
 =head1 SYNOPSIS
 
@@ -207,6 +217,16 @@ C<AnyEvent::WebSocket::Server> is ever created (after the
 handshake is complete, the client and server look pretty
 much the same).
 
+=head1 ATTRIBUTES
+
+=head2 handle
+
+The underlying L<AnyEvent::Handle> object used for the connection.
+WebSocket handshake MUST be already completed using this handle.
+You should not use the handle directly after creating L<AnyEvent::WebSocket::Connection> object.
+
+Usually only useful for creating server connections, see below.
+
 =head1 METHODS
 
 =head2 $connection-E<gt>send($message)
@@ -246,6 +266,14 @@ Called when the connection is terminated
 
 Close the connection.
 
+=head1 SERVER CONNECTIONS
+
+Although written originally to work with L<AnyEvent::WebSocket::Client>,
+this class was designed to be used for either client or server WebSocket
+connections.  For details, contact the author and/or take a look at the
+source for L<AnyEvent::WebSocket::Client> and the examples that come with
+L<Protocol::WebSocket>.
+
 =head1 DEPRECATED METHODS
 
 The methods in this section are deprecated and may be removed from a
